#include "CyanBus.hpp"

// C++ system level
#include <cstring>     // memset, strncmp
#include <cstdlib>     // TODO: is this for memcpy?
#include <cinttypes>
#include <functional>

// ESP32 specific
#include "esp_err.h"
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/uart.h"
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"

// project specific
#include "CyanBusCRC.hpp"
#include "LambdaTask.hpp"
#include "SpiderLib/Callback.hpp"
#include "SpiderLib/CallbackMap.hpp"

// qthing stuff
// #include ""



static const char* TAG = "CyanBus";


// Extract the length from a buffer; Network order / Big-Endian
static uint16_t getLength(const uint8_t* buffer) {
  return (buffer[0] << 8) | buffer[1];
}


namespace SiliconTorch {

  namespace CyanBus {

    // CyanBus envelope header
    const char* const HEADER = "fxCyan";


    CyanBus::CyanBus(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate, uint8_t uartChannel) : tx(tx), rx(rx), de(de), re(re), uartChannel(uartChannel) {

      uint64_t bitMask = 0;

      bitMask |= 1 << tx;
      bitMask |= 1 << de;
      bitMask |= 1 << re;

      gpio_config_t gpio = {
        .pin_bit_mask = bitMask,
        .mode         = GPIO_MODE_OUTPUT,
        .pull_up_en   = GPIO_PULLUP_DISABLE,
        .pull_down_en = GPIO_PULLDOWN_DISABLE,
        .intr_type    = GPIO_INTR_DISABLE,        
      };

      // configure outputs
      gpio_config(&gpio);
      txrxEN(false);

      gpio.pin_bit_mask = 1 << rx;
      gpio.mode = GPIO_MODE_INPUT;

      // configure inputs
      gpio_config(&gpio);


      uart_config_t uartConfig = {
        .baud_rate           = (int)baudRate,
        .data_bits           = UART_DATA_8_BITS,
        .parity              = UART_PARITY_DISABLE,
        .stop_bits           = UART_STOP_BITS_1,
        .flow_ctrl           = UART_HW_FLOWCTRL_DISABLE,
        .rx_flow_ctrl_thresh = 0,
        .use_ref_tick        = false,
        // .source_clk = UART_SCLK_APB,
      };


      uart_port_t _ch = (uart_port_t)uartChannel;

      ESP_ERROR_CHECK(uart_param_config(_ch, &uartConfig));
      ESP_ERROR_CHECK(uart_set_pin(_ch, tx, rx, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE));
      // uart_intr_config(_ch);

      ESP_ERROR_CHECK(uart_driver_install(_ch, MTU * 2, 0, 64, &uartEvQ, 0));  // txBuf = 0; queue size = 64

      ESP_ERROR_CHECK(uart_set_mode(_ch, UART_MODE_UART));

      // WARNING: RX timeout MUST be > 0 !!
      // It determines after how much symbols pause an event is raised (event queue uses TOUT-feature internally)
      uart_set_rx_timeout(_ch, 5);


      // TOOD: receive only ATM!
      txEN(false);
      rxEN(true);


      readerTaskHandle = new Util::LambdaTask([&](){ readerTask(); });

      // packetQ = xQueueCreate(3, sizeof(CyanData));
      packetQ = xQueueCreate(3, MTU);
      packetTaskHandle = new Util::LambdaTask([&](){ packetTask(); });
    }

    void CyanBus::readerTask() {

      uart_port_t _ch = (uart_port_t)uartChannel;

      const uint32_t bufLEN = MTU + 256;      // MTU + maxEventSize(120) + slack
      uint8_t* buffer = new uint8_t[bufLEN];  // our packet buffer
      uint8_t* bufPTR = buffer;               // points inside our buffer; used for envelope parsing…

      uint16_t payloadLEN = 0;
      const uint8_t crcLEN = 4;
      const uint8_t lengthLEN = 2;
      const uint8_t headerLEN = std::strlen(HEADER);

      std::memset(buffer, 0x00, bufLEN);


      uart_event_t event;

      auto Q = uartEvQ;
      auto Q2 = packetQ;

      while (true) {

        if (xQueueReceive(Q, (void*)&event, (TickType_t)portMAX_DELAY)) {  // 0 = return immediately!

          auto msgs  = uxQueueMessagesWaiting(Q);
          auto space = uxQueueSpacesAvailable(Q);

          size_t available;
          uart_get_buffered_data_len(_ch, &available);

          ESP_LOGD(TAG, "Got a new UART event! 🤩   QByte[ %d ]  |QMsg|[ %d ]  event.size[ %d ]  uart.available[ %d ]", space, msgs, event.size, available);

          switch (event.type) {

            // received plain data
            case UART_DATA: {

              if ((bufPTR - buffer) + event.size > bufLEN) {
                ESP_LOGD(TAG, "[W] |buffer|[ %d ] Overflow!", bufLEN);

                bufPTR = buffer;
                std::memset(buffer, 0x00, bufLEN);  // TODO: needed…?
                uart_flush_input(_ch);
                xQueueReset(Q);
                // fsm = SearchHeader;
                payloadLEN = 0;
              }


              bufPTR += uart_read_bytes(_ch, bufPTR, event.size, portMAX_DELAY);

              if (payloadLEN == 0 && bufPTR - buffer >= headerLEN + lengthLEN) {  // got enough for header/length parsing

                if (strncmp(reinterpret_cast<const char*>(buffer), HEADER, headerLEN) == 0) {
                  payloadLEN = getLength(buffer + headerLEN);
                  ESP_LOGD(TAG, "[D] Found header: |payload|[ %d ]", payloadLEN);

                  if (payloadLEN > MTU) {
                    ESP_LOGW(TAG, "|payload|[ %d ] exceeds MTU[ %d ]", payloadLEN, MTU);
                    bufPTR = buffer;
                    std::memset(buffer, 0x00, bufLEN);  // TODO: needed…?
                    uart_flush_input(_ch);
                    xQueueReset(Q);
                    // fsm = SearchHeader;
                    payloadLEN = 0;
                  }
                }
              }

              if (payloadLEN > 0 && (bufPTR - buffer) > headerLEN + payloadLEN + crcLEN) {  // packet complete!

                bool crcOK = Impl::checkCRC(buffer + headerLEN + lengthLEN, payloadLEN + crcLEN);

                char strbuf[17];
                strbuf[16] = 0x00;
                std::memcpy(strbuf, buffer, 16);

                ESP_LOGD(TAG, "PACKET DEBUG: Starts with header[ %s ], contains |data|[ %s ] and has crc⏻[ %s ]", strbuf, buffer + headerLEN + lengthLEN, crcOK ? "✔" : "✘");

                if (crcOK) xQueueSendToBack(Q2, (void*)(buffer + headerLEN), portMAX_DELAY);
                else ESP_LOGD(TAG, "[W] Packet CRC check failed!");

                bufPTR = buffer;
                std::memset(buffer, 0x00, bufLEN);  // TODO: needed…?    (if not: move to --> if not crcOK………)

                if (!crcOK) xQueueSendToBack(Q2, (void*)buffer, portMAX_DELAY);

                // TODO: do we really need this after successful packet reception? I think not!
                // uart_flush_input(_ch);
                // xQueueReset(Q);

                // fsm = SearchHeader;
                payloadLEN = 0;
              }

              break;
            }

            // got an RX break signal
            case UART_BREAK: {
              ESP_LOGD(TAG, "[I] Received BREAK 🛑 after reading bytes#[ %d ]", bufPTR - buffer);

              bufPTR = buffer;
              std::memset(buffer, 0x00, bufLEN);  // TODO: needed…?
              uart_flush_input(_ch);
              xQueueReset(Q);
              // fsm = SearchHeader;
              payloadLEN = 0;

              break;
            }

            default: {

              char* eventName;

              switch (event.type) {
                case UART_DATA:        { eventName = (char*)"UART_DATA";        break; }
                case UART_BREAK:       { eventName = (char*)"UART_BREAK";       break; }
                case UART_BUFFER_FULL: { eventName = (char*)"UART_BUFFER_FULL"; break; }
                case UART_FIFO_OVF:    { eventName = (char*)"UART_FIFO_OVF";    break; }
                case UART_FRAME_ERR:   { eventName = (char*)"UART_FRAME_ERR";   break; }
                case UART_PARITY_ERR:  { eventName = (char*)"UART_PARITY_ERR";  break; }
                case UART_DATA_BREAK:  { eventName = (char*)"UART_DATA_BREAK";  break; }
                case UART_PATTERN_DET: { eventName = (char*)"UART_PATTERN_DET"; break; }
                default:               { eventName = (char*)"_UNKNOWN_";        break; }
              }

              ESP_LOGD(TAG, "[W] Unexpected event[ %s ]!", eventName);

              bufPTR = buffer;
              std::memset(buffer, 0x00, bufLEN);  // TODO: needed…?
              uart_flush_input(_ch);
              xQueueReset(Q);
              // fsm = SearchHeader;
              payloadLEN = 0;
            }
          }


        } else {
          // we should never get here. let's see if this holds ;)
          ESP_LOGW(TAG, "Receiving from queue failed…?!");
        }
      }

      delete buffer;
    }

    void CyanBus::packetTask() {
      const uint8_t* buffer = new uint8_t[MTU];
      uint16_t payloadLEN = 0;

      auto Q = packetQ;
      while (true) {

        if (xQueueReceive(Q, (void*)buffer, (TickType_t)portMAX_DELAY)) {

          payloadLEN = getLength(buffer);
          bool crcOK = payloadLEN > 0;

          if (crcOK) ESP_LOGD(TAG, "[PacketTask]: |payload|[ %d ] starts with data[ %s ]", payloadLEN, buffer+2);
          else ESP_LOGD(TAG, "[PacketTask]: |payload|[ %d ] has incorrect crc⏻[ ✘ ]", payloadLEN);

          const PacketData packet = {
            .payload = buffer+2,
            .length = payloadLEN
          };

          const PacketMetrics metrics = {
            .crcOK = crcOK,
            .payloadLEN = payloadLEN
          };

          if (crcOK) packetCallback(buffer+2, payloadLEN, packet);
          metricsCallback(metrics);
        }
      }

      delete buffer;
    }


    void CyanBus::setBaudRate(uint32_t baudRate) {
      uart_set_baudrate((uart_port_t)uartChannel, baudRate);
    }

    uint32_t CyanBus::getBaudRate() {
      uint32_t baudRate = 0;
      uart_get_baudrate((uart_port_t)uartChannel, &baudRate);
      return baudRate;
    }


    void CyanBus::txEN(bool state) {
      gpio_set_level((gpio_num_t)de, state);
    }

    void CyanBus::rxEN(bool state) {
      gpio_set_level((gpio_num_t)re, state ^ 1);
    }
    
    void CyanBus::txrxEN(bool state) {
      txEN(state);
      rxEN(state);
    }

  }
}
